# kilibs is free software: you can redistribute it and/or modify it under the terms of
# the GNU General Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# kilibs is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with kilibs.
# If not, see < http://www.gnu.org/licenses/ >.
#
# (C) The KiCad Librarian Team

from __future__ import annotations

from KicadModTree.nodes.base import Arc
from KicadModTree.nodes.NodeShape import NodeShape
from KicadModTree.util.line_style import LineStyle
from kilibs.geom import GeomArc, GeomCompoundPolygon, GeomLine, GeomPolygon, Vector2D


class CompoundPolygon(NodeShape, GeomCompoundPolygon):
    def __init__(
        self,
        shape: (
            CompoundPolygon
            | GeomCompoundPolygon
            | list[Vector2D | GeomPolygon | GeomLine | GeomArc]
            | None
        ) = None,
        layer: str = "F.SilkS",
        width: float | None = None,
        style: LineStyle = LineStyle.SOLID,
        fill: bool = False,
        offset: float = 0,
        serialize_as_fp_poly: bool = True,
        close: bool = True,
    ):
        """Create a geometric compound polygon.

        Args:
            shape: compound polygon, list of points, polygons, lines or arcs  from which
                to derive the compound polygon.
            layer: Layer.
            width: Line width in mm. If `None`, then the standard width for the given
                layer will be used when the serializing the node.
            style: Line style.
            fill: `True` if the compound polygon is filled, `False` if only the outline
                is visible.
            offset: Amount by which the compound polygon is inflated or deflated (if
                offset is negative).
            serialize_as_fp_poly: If True, will serialize the compound polygon as
                fp_poly, leading to line-segment-appromimations of contained arcs in the
                FP editor (the file format already supports arcs, but the editor does
                not) if False, will break the polygon into free primitives, yielding
                true arcs in the FP editor but no single primitive. Grouping will be
                used.
            close: If `True` the polygon will form a closed shape. If `False` there
                won't be any connecting line between the last and the first point.
        """
        self.init_super(kwargs=locals())

    def getVirtualChilds(self) -> list[NodeShape]:
        """Return a list containing the child nodes."""
        if self.serialize_as_fp_poly and self.close:
            return []
        else:
            return self.to_child_nodes(list(self.get_atomic_shapes()))

    def get_fp_poly_elements(self) -> list[Vector2D | Arc]:
        points_and_arcs: list[Vector2D | Arc] = []
        for geom in self.get_points_and_arcs():
            if isinstance(geom, Vector2D):
                points_and_arcs.append(geom)
            else:
                arc = Arc(
                    shape=geom, layer=self.layer, width=self.width, style=self.style
                )
                arc._parent = self._parent
                points_and_arcs.append(arc)
        return points_and_arcs
